/* An Implementation of Floyd's All-Pairs Shortest Path Algorithm
 * ----------------------------------------------------------------------------
 * Author:  Shane Saunders
 */
#include <stdlib.h>  /* for malloc() and free() */
#include <limits.h>  /* for LONG_MAX */
#include <stdio.h>   /* for text output */

/* Type definition for shortest path distances. */
typedef long distance;

/* Infinite Distance */
const distance INFINITE_DIST = LONG_MAX/2;

/* --- alloc_matrix() ---
 * Allocate an n by n matrix.
 */
distance **alloc_matrix(int n) {
    int i;
    distance **M = malloc(n * sizeof(distance *));
    for(i = 0; i < n; i++) {
        M[i] = malloc(n * sizeof(distance));
    }
    return M;
}

/* --- free_matrix() ---
 * Free the space used by a matrix.
 */
void free_matrix(distance **M, int n) {
    int i;
    for(i = 0; i < n; i++) {
        free(M[i]);
    }
    free(M);
}

/* --- urand() ---
 * Return a uniformly distributed random real number between 0 and 1.
 */
double urand() {
    return (double)rand()/RAND_MAX;
}

/* --- create_random_matrix() ---
 * Create a random distance matrix.  Here prob is the probability of edge
 * existence, dmin is the minimum distance, and dmax is the maximum
 * distance (inclusive).
 */
distance **create_random_matrix(int n, double prob,
                                distance dmin, distance dmax)
{
    int i,j;
    distance drange = 1 + dmax - dmin;
    distance **M = alloc_matrix(n);
    
    /* create random edges according to probability prob */
    for(i = 0; i < n; i++) {
        for(j = 0; j < n; j++) {
            M[i][j] = urand() < prob ? dmin + rand() % drange : INFINITE_DIST;
        }
    }
    return M;
}

/* --- dump_matrix() ---
 * Print a text representation of the matrix to the standard output
 */
void dump_matrix(distance **M, int n) {
    int i, j;
    for(i = 0; i < n; i++) {
        for(j = 0; j < n; j++) {
            if(M[i][j] == INFINITE_DIST) {
                printf("   inf");
            }
            else {
                printf(" %5ld", M[i][j]);
            }
        }
        printf("\n");
    }
}

/* --- Floyd's algorithm ---
 * The parameter M points to a distance matrix of size n, representing the
 * edges of the graph.  Running Floyd's algorithm will convert these into the
 * all-pairs shortest path distances.
 */
void floyd(distance **M, int n) {
    int i, j, k;
    for(k = 0; k < n; k++) {
        for(i = 0; i < n; i++) {
            for(j = 0; j < n; j++) {
                if(M[i][j] > M[i][k] + M[k][j]) M[i][j] = M[i][k] + M[k][j];
            }
        }
    }
}

#define SHOW_OUTPUT 0

int main(void)
{
    const int nVertices = 1000;
    distance **M;
    
    /* create a random distance matrix */
    M = create_random_matrix(nVertices, 0.5, 1, 10000);
#if SHOW_OUTPUT
    printf("Distance Matrix:\n");
    dump_matrix(M, nVertices);
#endif
    
    /* run Floyd's algorithm */
    floyd(M, nVertices);
#if SHOW_OUTPUT
    printf("Shortest Path Distances:\n");    
    dump_matrix(M, nVertices);    
#endif
    
    /* tidy up */
    free_matrix(M, nVertices);
        
    return 0;
}
