/***  File: mst.c - Minimal Spanning Tree Algorithms ***/
/*
 *    Shane Saunders
 */
#include <stdlib.h>
#include "../graphs/graph.h"
#include "mst.h"

/*** Special values. ***/
#define TRUE 1
#define FALSE 0

/*--- MST Algorithms --------------------------------------------------------*/

/*** Prim's Algorithm ***/

/* mst_prim() - Minimum spanning tree algorithm.  Computes the minimum spanning
 * tree of the graph pointed to by g.  Returns the result as a prim_result_t
 * structure (see prim.h).
 */
graph_t *mst_prim(const graph_t *g, int v0, const heap_info_t *heap_info)
{
    int v, w;
    int *s;
    long dist, *d;
    int n;
    
    graph_vertex_t *vertices;
    graph_edge_t *edge_ptr;
    
    void *front;
    int (*heap_delete_min)(void *);
    void (*heap_insert)(void *, int, long);
    void (*heap_decrease_key)(void *, int, long);
    int (*heap_n)(void *);
    void *(*heap_alloc)(int);
    void (*heap_free)(void *);
    void (*heap_dump)(void *);
    
    graph_t *result;
    int *pred;


    /* Set up pointers to heap interface functions. */
    heap_delete_min = heap_info->delete_min;
    heap_insert = heap_info->insert;
    heap_decrease_key = heap_info->decrease_key;
    heap_n = heap_info->n;
    heap_alloc = heap_info->alloc;
    heap_free = heap_info->free;
    heap_dump = heap_info->dump;

    /* Provide quick access to graph fields. */
    vertices = g->vertices;
    n = g->n_vertices;

    /* Allocate distance and predecessor arrays. */
    pred = malloc(n * sizeof(int));
    d = calloc(n, sizeof(long));
    
    /* Allocate arrays used by the algorithm. */
    s = calloc(n, sizeof(int));
    
    /* Create the heap. */
    front = heap_alloc(n);

    /* Initially all vertices are unreached. */
    for(v = 0; v < n; v++) pred[v] = -1;

    /* The start vertex is part of the solution set. */
    s[v0] = TRUE;
    d[v0] = 0;
    
    /* Put out set of the starting vertex into the frontier and update the
     * distances to vertices in the out set.  k is the index for the out set.
     */
    edge_ptr = vertices[v0].head_edge;
    while(edge_ptr) {
        w = edge_ptr->target;
        dist = d[w] = edge_ptr->dist;
        heap_insert(front, w, dist);
/*         printf("inserted %d(%d)\n", w, dist);  heap_dump(front); */
	pred[w] = v0;
	edge_ptr = edge_ptr->next;
    }

    /* At this point we are assuming that all vertices are reachable from the
     * starting vertex and N > 1 so that j > 0.
     */

    while(heap_n(front) > 0) {

        /* Find the vertex in frontier that has minimum distance. */
        v = heap_delete_min(front);
/*         printf("deleted min %d\n", v);  heap_dump(front); */

        /* Add the vertex to the solution set. */
        s[v] = TRUE;

        /* Update distances to vertices, w, in the out set of v.
         */
        edge_ptr = vertices[v].head_edge;
	while(edge_ptr) {
	    w = edge_ptr->target;

            /* Only update if w is not already in the solution set.
             */
            if(!s[w]) {

                /* If w is in the frontier (pred[w] >= 0) the new edge
		 * distance to w is the minimum of its current distance and the
		 * edge distance (v,w).
                 */
                dist = edge_ptr->dist;
                if(pred[w] >= 0) {
                    if(dist < d[w]) {
                        d[w] = dist;
                        heap_decrease_key(front, w, dist);
/*                         printf("decreased %d(%d)\n", w, dist);  heap_dump(front); */
			pred[w] = v;
                    }
                }
                else {
                    d[w] = dist;
                    heap_insert(front, w, dist);
/*                     printf("inserted %d(%d)\n", w, dist);  heap_dump(front); */
                    pred[w] = v;
                }
            } /* if */

	    edge_ptr = edge_ptr->next;
        } /* while */
    } /* while */

    /* Assemble the result graph (vertex zero is the source) */
    result = graph_create_blank(n);
    for(v = 1; v < n; v++) {
        graph_add_edge(result, pred[v], v, d[v]);
    }

    /* Free space used by arrays local to this function. */
    free(s);
    heap_free(front);
    free(pred);
    free(d);

    return result;
}


/*** Kruskal's algorithm ***/

/* Quicksort algorithm used by Kruskal's algorithm ***/
void mst_kruskal_sort(graph_edge_t **p, graph_edge_t **q);

/* mst_kruskal() - Minimum spanning tree algorithm.  Computes the minimum
 * spanning tree of the undirected graph pointed to by g.  Returns the result
 * as a prim_result_t structure (see prim.h).
 */
graph_t *mst_kruskal(const graph_t *g) {
    int i, j, n, m, v, w, vs,ws;
    graph_vertex_t *vertices;
    graph_edge_t *edge_ptr;
    graph_edge_t **edges;
    graph_t *result;
    int *owner, *size;

    /* Initialise variables */
    n = g->n_vertices;
    m = g->n_edges;
    vertices = g->vertices;

    /* Allocate the result graph. */
    result = graph_create_blank(n);
    
    /* Allocate arrays used by this algorithm. */
    edges = malloc(m*sizeof(graph_edge_t *));
    owner = malloc(n*sizeof(int));
    size = malloc(n*sizeof(int));
    
    /* Initialise the edge array to contain a pointer for every edge in the
     * graph.  Reverse edges are skipped by enumerating only those edges
     * v-->w for which v > w.
     */
    j = 0;
    for(v = 0; v < n; v++) {
        edge_ptr = vertices[v].head_edge;
	while(edge_ptr) {
            if(edge_ptr->target > v) edges[j++] = edge_ptr;
	    edge_ptr = edge_ptr->next;
	}
    }

    /* Use quicksort to sort the array of j edges in ascending order of
     * distance.
     */
    mst_kruskal_sort(&edges[0], &edges[j-1]);

    /* The initial spanning forest contains n 1-vertex trees. */
    for(i = 0; i < n; i++) owner[i] = i;
    for(i = 0; i < n; i++) size[i] = 1;
    
    /* Gradually grow the forest until a single spanning tree is left.
     * Three tree number of a vertex is the number of the root vertex of the
     * tree in which it belongs.
     */
    for(i = 0; i < j; i++) {
        edge_ptr = edges[i];
	v = edge_ptr->reverse->target;
	w = edge_ptr->target;

        /* find the sets containing v and w */
        vs = v;
        while(vs != owner[vs]) vs = owner[vs];
        ws = w;
        while(ws != owner[ws]) ws = owner[ws];
        
	/* If v and w are in different sets, then the edge joins the two trees
         * of v and w, and the sets merge.
	 */
      	if(vs != ws) {
            /* add edge to result */
            graph_add_edge(result, v, w, edge_ptr->dist);

            /* merge sets */
            if(size[vs] >= size[ws]) {
                owner[ws] = vs;
                size[vs] += size[ws];
            }
            else {
                owner[vs] = ws;
                size[ws] += size[vs];
            }
	}
    }

    return result;
}

/* mst_kruskal_sort() - Performs quicksort recursively on the section of
 * the array specified by pointers p and q.  Here p and q are pointers to edge
 * pointers.
 */
graph_edge_t *x;  /* pivot used in quicksort */
int xdist;
void mst_kruskal_sort(graph_edge_t **p, graph_edge_t **q)
{
    graph_edge_t **i, **j;

    /* Sorting is only required if p < q; that is, two ore more elements are
     * being sorted.
     */
    if(p < q) {

        /* Select the pivot, x, as first entry in the array region. */
	x = *p;
        xdist = x->dist;

	/* Proceed with partitioning, which places all elements smaller than x
	 * to x's left, and all elements greater than x to x's right.
	 */
	
	i = p;
	j = q + 1;

	while(i < j) {
	    
	    do {
		j--;
		if(i == j) goto partition_end;
	    } while((*j)->dist > xdist);  /* Better balance using > rather
					   * than >=.  (can be much faster)
					   */
	    *i = *j;

	    do {
		i++;
		if(i == j) goto partition_end;
	    } while((*i)->dist < xdist);  /* Better balance using < rather
					   * than <=.  (can be much faster)
					   */
	    *j = *i;
	}
    
      partition_end:
	*i = x;

	/* Now perform quicksort on the region to the left of x, and the
	 * region to the right of x.
	 */
	mst_kruskal_sort(p, i - 1);
	mst_kruskal_sort(i + 1, q);
    }
}

