/*** mst_test.c - Test minimal spanning tree algorithms. ***/
/*
 *   Shane Saunders
 */
#include <stdio.h>
#include <stdlib.h>
#include "mst.h"
#include "../graphs/graph.h"
#include "../timing/timing.h"

/* #include all heaps to be tested using Prim's algorithm. */
#include "../heaps/bheap.h"
#include "../heaps/fheap.h"
#include "../heaps/ttheap.h"
#include "../heaps/triheap.h"
#include "../heaps/triheap_ext.h"


/* Step size between values of n used. */
#define STEP 10000
#define NMAX 100000

/*** Main Program ***/
int main(int argc, char *argv[])
{
    int i, k, n_max, n_samples;
    double edge_f;
    graph_t *r;
    graph_t *graph;
    timing_t *t;

    /* Process command line arguments (if any), otherwise get input from the
     * user.  Arguments supplied correspond to:
     *     - number of samples used for average calculation.
     *     - maximum graph size to test.
     *     - edge factor (the average size of the out set).
     */
    if(argc == 3) {
        n_samples = atoi(argv[1]);
        edge_f = atof(argv[2]);
    }
    else {

	/* Several samples may be needed to calculate the time to the required
	 * accuracy.
	 */
	printf("Enter the number samples to use: ");
	scanf("%d", &n_samples);

	/* The edge factor entered should be greater than 1 - 1/n
	 */
	printf("Enter the edge factor (i.e. average size of OUT set): ");
	scanf("%lf", &edge_f);
	putchar('\n');
    }
    n_max = NMAX;

    /* Allocate structure for timing information. */
    t = timing_alloc(2);

    printf("CPU Time (msec) for Spanning Tree Algorithms\n");
    printf("Graph Size = n, Edge factor = %.3f, Number of samples = %d\n", edge_f, n_samples);

    /* Print column labels */
    printf("\nResults:\nn\tPrim\tKruskal\n");

    
    /* Test over varying graph sizes. */
    for(k = STEP; k <= n_max; k += STEP) {

	/* Initialise the sums used in average calculations to zero. */
        timing_reset(t);

	/* For the average calculation, sum the number of comparisons. */
	for(i = 0; i < n_samples; i++) {

	    /* We use a new random graph for each run of Prim's algorithm.
	     */
            graph = random_graph(k, edge_f*k);
            graph_check(graph);

	    /* Time the Prim and Kruskal algorithms. */
	    /* The heap_info structure of each heap is passed to Prim's
	     * algorithm in order to use the functions provided by that heap.
	     */
  	    timing_start();
            r = mst_prim(graph, 0, &BHEAP_info);
/*                 graph_dump(graph); */
/*                 graph_dump(r); */                
            timing_stop(t,0);
            graph_free(r);
            timing_start();
            r = mst_kruskal(graph);
            timing_stop(t,1);
            graph_free(r);
	    
	    /* Free the random graph. */
            graph_free(graph);
	}

	/* Print line of key comparison results for current value of k. */
	printf("%d", k);  timing_print(t,"\t%.0f",n_samples);  putchar('\n');
    }

    return 0;
}
